/*
 *
 * Author           : Ivan Barrera (Bruce) <Bruce@Ivn.cl>
 * Updated          : 2007.04.06
 * Status           : Functional
 * Current Version  : v0.2
 *
 * NAME
 *     mod_vhost_limit.c
 *
 * PURPOSE
 *       * Restrict the number of simultaneous connections per vhost
 *
 * HISTORY
 *     2007.04.06
 *       Included some cleanup, and fixes suggested (and coded by) Chirouze Olivier
 *     2004.02.24
 *       Initial revision 0.1
 *
 * IMPORTANT
 *
 * - This program source code, is free to distribute and use personally. If you 
 *  like the work i did porting it to Apache 2 , and the new features i'll be
 *  implementing on the future, donations will be well received. Im a programmer
 *  who is just starting to code for life, so ill need lots of support.
 *   How could you cooperate ? I live in Chile. Money orders are good idea,
 *  but with high cost. I also have paypal, but i cant withdraw money from here.
 *   Postals, a bill (to know money from other countries), or better yet, some
 *  job to do online.
 *
 * - All the software i'll be developing, will be under the same license. Not yet
 *  defined, but as soon as its ready, i'll update this files.
 *
 * - If you use this software commercially, please let me know. A donation is good
 *  as with it, i'll be willing to give support, and work more on this modules.
 *
 * COMMENTS 
 *
 *       If ExtendedStatus is Off, it wont work. Anyways, you'll be warned
 *         I was based on mod_vhost_limit from Takato Satsuma <takato@nowhere-land.org>
 *       who wrote it for Apache1.3.
 *
 *         If you run into problems, be sure that ExtendedStatus, is set On before the
 *       LoadModule sentence. 
 *
 *
 */

#define VERSION             0.2
#define CORE_PRIVATE
#define VHOST_LIMIT_DEBUG   0

#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_main.h"
#include "ap_mpm.h"
#include "http_log.h"
#include "scoreboard.h"

int server_limit, thread_limit;
module AP_MODULE_DECLARE_DATA vhost_limit_module;

typedef enum {
    VHOST_LIMIT_UNSET,
    VHOST_LIMIT_SET
} vhost_limit_mode;

typedef struct vhost_limit_conf {
    int max_clients;
    vhost_limit_mode mode;
} vhost_limit_conf;

static const char *
max_vhost_clients_set(cmd_parms *cmd, void *dummy, const char *arg)
{
    int val;
    vhost_limit_conf *conf;

    if (!ap_extended_status)
        return "ExtendedStatus must be set to On for mod_vhost_limit to be available";

    val = atoi(arg);
    conf = (vhost_limit_conf *) ap_get_module_config(cmd->server->module_config, &vhost_limit_module);

    if (val < 1)
        return "MaxVhostClients must be an integer greater than 0";
    else {
        conf->max_clients = val;
        conf->mode = VHOST_LIMIT_SET;
#if VHOST_LIMIT_DEBUG
        ap_log_error(APLOG_MARK,APLOG_NOTICE|APLOG_NOERRNO, 0 ,NULL,
                     "mod_vhost_limit: %s MaxVhostClients = %d",cmd->server->server_hostname, val);
#endif
    }

    return NULL;
}

static const char *check_extendedstatus(cmd_parms *cmd, void *dummy, int arg)
{
    const char *err = ap_check_cmd_context(cmd, GLOBAL_ONLY);
    if (err != NULL) {
        return err;
    }
    ap_extended_status = arg;
    return NULL;
}

static const command_rec vhost_limit_cmds[] =
{
    AP_INIT_FLAG ( "ExtendedStatus" , check_extendedstatus , NULL, RSRC_CONF, "\"On\" to enable extended status information, \"Off\" to disable"),
    AP_INIT_TAKE1( "MaxVhostClients", max_vhost_clients_set, NULL, RSRC_CONF, "'MaxVhostClients' followed by non-negative integer"),
    {NULL}
};


static int vhost_limit_handler(request_rec *r)
{
    int j, i, matched=0;
    worker_score *ws_record;
    process_score *ps_record;
    
    vhost_limit_conf *conf;
    // Changed size to 32 because Apache vhosts are stored in 32 chars!!!
    // If we want the comparison to work, we need to store it in 32 too.
    char cut_req_host[31] = "";
    const char *req_host;
    const char *vhost;
    
    conf = (vhost_limit_conf *) ap_get_module_config(r->server->module_config,&vhost_limit_module);
    
    if (conf->mode == VHOST_LIMIT_UNSET) 
        return DECLINED; 

    if (!ap_exists_scoreboard_image()) {
        ap_log_error(APLOG_MARK,APLOG_NOTICE|APLOG_NOERRNO, 0 ,NULL, "mod_vhost_limit unavailable in inetd mode");
        return HTTP_INTERNAL_SERVER_ERROR;
    }

 
   /********************************************
    *  Vhost Counting - Start                  *
    ********************************************/

    req_host = ap_get_server_name(r);
    
    
    for (i = 0; i < server_limit; ++i) 
    {
        for (j = 0; j < thread_limit; ++j) 
        {
            ps_record = ap_get_scoreboard_process(i);
            ws_record = ap_get_scoreboard_worker(i, j);
            vhost = ws_record->vhost;

            if (ps_record->generation != ap_my_generation)
                vhost = NULL;
            if (!vhost)
                continue;
    
            // Changed size to 32 because Apache vhosts are stored in 32 chars!!!
            // If we want the comparison to work, we need to store it in 32 too.
            strncpy(cut_req_host, req_host, 31);
     
    
            if ((strcmp((char*)(&cut_req_host), vhost)==0) &&
                (ws_record->status == SERVER_BUSY_READ || ws_record->status == SERVER_BUSY_WRITE ||
                 ws_record->status == SERVER_BUSY_LOG  || ws_record->status == SERVER_BUSY_DNS ||
                 ws_record->status == SERVER_BUSY_KEEPALIVE)) 
            {

                //short slot_predates = 1;

                switch (ws_record->status) {
                    case SERVER_BUSY_KEEPALIVE:
                    case SERVER_BUSY_READ:
                        matched++;
                        break;
                     case SERVER_BUSY_WRITE:
                     case SERVER_BUSY_LOG:
                     case SERVER_BUSY_DNS:
                        if (ws_record->conn_count > 1) 
                            matched++;
                        else
                            matched++;
                        break;
                     default:
            break;
                } /* switch */
            } /* if ((strcmp(req_host,vhost)==0) */
            
            
        } /* for (j...) */
    } /* for (i...) */

   /********************************************
    *  Vhost Counting - End                    *
    ********************************************/

    if (matched > conf->max_clients) 
    {
        ap_log_error(APLOG_MARK, APLOG_WARNING, 0,NULL,
                     "Vhost Limit : Access to %s deferred, Max Clients %d exceeded (%d currently)", 
                     req_host, conf->max_clients, matched);
        return HTTP_SERVICE_UNAVAILABLE;
    }
#if VHOST_LIMIT_DEBUG
    else
    {
        ap_log_error(APLOG_MARK, APLOG_NOTICE|APLOG_NOERRNO, 0,NULL,
                     "Vhost Limit %s : Max Clients %d (%d currently)",
                     req_host, conf->max_clients, matched);
    }
#endif

    return DECLINED;
}

static void *create_vhost_limit_config(apr_pool_t *p, server_rec *s)
{
    vhost_limit_conf *conf;

    conf = (vhost_limit_conf *) apr_pcalloc(p, sizeof(vhost_limit_conf));
    conf->max_clients = server_limit;
    conf->mode = VHOST_LIMIT_UNSET;

    return conf;
}

static int vhost_limit_init(apr_pool_t *p, apr_pool_t *plog, apr_pool_t *ptemp,
                       server_rec *s)
{

    ap_mpm_query(AP_MPMQ_HARD_LIMIT_THREADS, &thread_limit);
    ap_mpm_query(AP_MPMQ_HARD_LIMIT_DAEMONS, &server_limit);
    return OK;
}

static void register_hooks(apr_pool_t *p)
{
    ap_hook_handler(vhost_limit_handler, NULL, NULL, APR_HOOK_REALLY_FIRST);//APR_HOOK_MIDDLE);
    ap_hook_post_config(vhost_limit_init, NULL, NULL, APR_HOOK_REALLY_FIRST);//APR_HOOK_MIDDLE);
}

module AP_MODULE_DECLARE_DATA vhost_limit_module =
{
    STANDARD20_MODULE_STUFF,
    NULL,                       /* dir config creater */
    NULL,                       /* dir merger --- default is to override */
    create_vhost_limit_config,  /* server config */
    NULL,                       /* merge server config */
    vhost_limit_cmds,           /* command table */
    register_hooks              /* register_hooks */
};


